--[[

    File: hopfield-gpu.lua 

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

--]]

--[[
Hopfield...
start with a vector of pixels ...
we're gonna be feeding it through a matrix,
so our texture area better be <= our max texture dimension size

let's assume 2048 is our max texture size = 2^11,
then our pictures are 2^5 x 2^6 at largest

for simplicity let's just say 2^5 and 1024
... I think that restriction applies to floating point textures on my card anyways
so we downsample all our input images to 2^5 x 2^5

then we initialize the network...
W = 1/n sum from i=1 to m of (v[i] outer v[i]) 
where v[i] is the i'th input pattern vector v in n-dimensional space
and I don't know if m <= n or not ...


--]]

if not glush then require 'glush' end
require 'common.pingpong'
require 'common.shader'

glush.gui.checkbox('doUpdate', false)
glush.gui.checkbox('doSingleUpdate', false)
glush.gui.checkbox('doReset', false)
glush.gui.slider('timestep', 0, 1, 1)
glush.gui.slider('pointSize', 1, 1, 10)
glush.gui.checkbox('erase',false)

-- TODO - make this editable
-- how about an editable image gui?
-- or a color gui with a feedback? (think fluid sim)
--initialState = glush.createTex2D('hopfield/initialstate.jpg')
--glush.gui.tex2D('initialState')


--[[
something to think about:
right now it's operating on 4 32x32 image sizes, using four 1024x1024 matrix to store them all
for starters we don't want four separate weight sets.  that would allow one channel to converge on one image while another does on another image.

so we want to turn this into a 32x32x4 image with a 4096x4096 matrix to store it all
that exceeds our maximum floating point texture size
we could circumvent that though with multiple matrices and a composite matrix multiply function...
each 1024x1024 float matrix ... rgba becomes 4096x1024 matrix ... takes up 4x as many bytes ... is 16m in memory
four of those would mean 64m of graphics memory
which sounds feasable... but i think will push my card's limits
we could stack four to make a 4kx4k matrix
--]]
local sampleDim = 32 	-- stored images can be up to 32x32 in size
local sampleDimSq = sampleDim * sampleDim	-- relations between all image pixels to themselves is (32*32)*(32*32) = 1024*1024

local genSampleShader = KernelShader.new{
	texs={'tex'};
	code=[[
void main() {
	const vec4 greyscale = vec4(.3, .6, .1, 0.);
	vec4 color = texture2D(tex, pos);
	float grey = dot(color, greyscale);
	//grey = 1. - step(grey, .5);
	gl_FragColor = vec4(grey,grey,grey,1.);
}
]];
}

local signShader = KernelShader.new{
	texs={'tex'};
	code=[[
void main() {
	vec4 c = texture2D(tex, pos);
	c = c * 2. - 1.;
	c = sign(c);
	c = c * .5 + .5;
	gl_FragColor = c;
}
]];
}

--[[
src is the source weight (size DIMSQxDIMSQ)
pattern is the new sample (size DIMxDIM)
I'm gonna store my weights as [-1,1] so I don't need a bias
that means for all my [0,1] incoming color values, make sure to *2-1
--]]
local initWeightShader = KernelShader.new{
	texs={'weights', 'pattern'};
	uniforms={patternScale='float'};
	code=
'#define DIMSQ '..sampleDimSq..'.\n'..
'#define DIM '..sampleDim..'.\n'..
[[
/*
interpolate
from src
to unravelled(pattern) outer unravelled(pattern)
by a factor of 1/n
for the n'th iteration 1,2,3,...
*/	

void main() {
	vec4 srcWeight = texture2D(weights, pos);

	//here we go from 'pos', in (x,y) resolution DIMSQ,DIMSQ
	//to 'patternPos' in x,y resolution DIM,DIM
	vec2 posi = pos * DIMSQ - .5;	//posi is 0...DIMSQ-1, 0...DIMSQ-1
	vec2 yis = floor(posi / DIM);	//get the integer-range y coordinates of our two samples
	vec2 xis = posi - DIM * yis;	//get the integer-range x coordinates of our two samples
	//transpose the 2x2 vector to get our integer-based sample1 and sample2
	vec2 pi1 = vec2(xis[0], yis[0]);
	vec2 pi2 = vec2(xis[1], yis[1]);
	//convert back to texcoords
	vec2 p1 = (pi1+.5)/DIM;
	vec2 p2 = (pi2+.5)/DIM;
	//and use that for our outer product lookup indexes
	//all in greyscale unfortunately...
	//because any more than that and I'd have to make a bigger matrix multiplier...
	//which I might end up doing...
	vec4 color1 = texture2D(pattern, p1) * 2. - 1.;
	vec4 color2 = texture2D(pattern, p2) * 2. - 1.;
	vec4 patternWeight = color1 * color2;	//some say divide by the input vector size

	gl_FragColor = srcWeight + patternWeight * patternScale;
#if 0
	//we're supposed to zero the weights along the diagonal...
	//so some implementations say...
	//(but in linear dynamic systems, this eliminates a positive trace	-
	//	- and positive traces (with positive determinants) turn fixed points into fixed sources,
	//	- which would keep our state vector from converging on a memory, and would not be good)
	//but Wolfram doesn't say to do this ...
	float del = pos.x - pos.y;
	del *= del;
	del *= 10000.;
	del = clamp(del, 0., 1.);
	gl_FragColor = gl_FragColor * del + .5 * (1. - del);
#endif
}
]];
}

local fbo = FBO.new()

local weights = PingPong.new{
	internalFormat=gl.GL_RGBA_FLOAT32;
	width=sampleDimSq;
	height=sampleDimSq;
	data=function() return 0,0,0,0 end;
	minFilter=gl.GL_NEAREST;
	magFilter=gl.GL_NEAREST;
}

local sampleTexs = table.new()

-- NOTICE: this takes tex for its own, so don't modify it afterwards
function addSample(tex, scale)
	local sampleTex = Tex2D.new{
		width=sampleDim;
		height=sampleDim;
		minFilter=gl.GL_NEAREST;
		magFilter=gl.GL_NEAREST;
	}

	fbo:draw{
		dest=sampleTex;
		viewport={0,0,sampleDim,sampleDim};
		shader=genSampleShader;
		resetProjection=true;
		texs={tex};
	}
	sampleTexs:insert{sample=sampleTex, full=tex}

	-- use them to generate the weights
	weights:swap()
	weights:draw{
		viewport={0,0,sampleDimSq, sampleDimSq};
		resetProjection=true;
		texs={weights:prev(), sampleTex};
		shader=initWeightShader;
		uniforms={patternScale=scale};
	}
end


local loadFromPatternFile = true
local loadFromFontFile = false

--[[
-- TODO - if patterns.txt exists then lfs.dir each pattern and use it
-- if it doesn't exist then lfs.dir *.jpg *.jpeg *.png *.bmp *.tiff *.tga
--]]
if loadFromPatternFile then
	local patterns = {}
	for line in io.lines('hopfield/patterns.txt') do
		table.insert(patterns, Tex2D.load('hopfield/'..line))
	end
	for _,v in ipairs(patterns) do
		addSample(v, 1/#patterns)
	end
end

if loadFromFontFile then
	local font = Tex2D.load('consoledata/font.png')
	--for index=0,255 do
	local letters = {}	--space?
	for index=string.byte('a')-32,string.byte('g')-32 do
		table.insert(letters, index)
	end
	for _,index in ipairs(letters) do
		local i = (index%16)+1
		local j = math.floor(index/16)+1

		-- collect the sampled-down textures
		local charDim = 16
		local tex = Tex2D.new{
			width=charDim;
			height=charDim;
			minFilter=gl.GL_NEAREST;
			magFilter=gl.GL_NEAREST;
		}
		
		Tex2D:enable()
		fbo:draw{
			dest=tex;
			viewport={0,0,charDim,charDim};
			resetProjection=true;
			texs={font};
			callback=function()
				gl.glBegin(gl.GL_QUADS)
				gl.glTexCoord(	(i-31/32)/16,	(j-31/32)/16	)	gl.glVertex(0,0)
				gl.glTexCoord(	(i-1/32)/16,	(j-31/32)/16	)	gl.glVertex(1,0)
				gl.glTexCoord(	(i-1/32)/16,	(j-1/32)/16	)	gl.glVertex(1,1)
				gl.glTexCoord(	(i-31/32)/16,	(j-31/32/16)/16	)	gl.glVertex(0,1)
				gl.glEnd()
			end;
		}
			
		addSample(tex, 1/#letters)
	end
	--font:delete()
end

weights:swap()
weights:draw{
	resetProjection=true;
	viewport={0,0,weights.width,weights.height};
	shader=signShader;
	texs={weights:prev()};
}

-- once all samples are loaded, init the state

-- the state is gonna be displayed, and is typically binary-clamped anyways, so who cares if the format wanes a bit?
local state = PingPong.new{
	internalFormat=gl.GL_RGBA_FLOAT32;
	width=sampleDim;
	height=sampleDim;
	minFilter=gl.GL_NEAREST;
	magFilter=gl.GL_NEAREST;
}

--here's our init. run this when the user changes the initialState texture.  i need a callback for that.
function resetState()
	gl.glEnable(gl.GL_TEXTURE_2D)
	state:draw{
		viewport={0,0,sampleDim,sampleDim};
		resetProjection=true;
		color={0,0,0};
	}
end
resetState()


-- the for loop goes very slow
-- better use a reduction shader...
local convergeShader = KernelShader.new{
	texs={'weights', 'state'};
	uniforms={dt='float'};
	code=
'#define DIMSQ '..sampleDimSq..'.\n'..
'#define DIM '..sampleDim..'.\n'..
[[
//wolfram says dx/dt = -x + W * s(x) for euler integrator
//which translates to (x(t+dt)-x(t))/dt = -x(t) + W * s(x(t))
// which becomes x(t+dt) = lerp from x(t) to W*s(x(t)) by a factor of dt

#define tanhscale 	1.
#define tanh(x) (2. / (1. + exp(-2. * tanhscale * (x))) - 1.)
#define signal(x)	tanh(x)
//#define signal(x) (1. - step(x, 0.) * 2.)
//#define signal(x)	sign(x)

void main() {
	vec4 src = texture2D(state, pos) * 2. - 1.;
	//our destination is 'pos'
	//unravel that to get our matrix column
	vec2 posi = pos * DIM - .5;
	float coli = posi.x + posi.y * DIM;
	float col = (coli + .5) / DIMSQ;
	vec2 statetci;
	vec4 accum = vec4(0.,0.,0.,0.);
	for (statetci.x = 0.; statetci.x < DIM; statetci.x+=1.) {
		for (statetci.y = 0.; statetci.y < DIM; statetci.y+=1.) {
			//our state vector lookup
			vec2 statetc = (statetci + .5) / DIM;
			//state color.  don't forget to remap from [0,1] to [-1,1]
			vec4 statecolor = texture2D(state, statetc) * 2. - 1.;
			//statecolor = signal(statecolor);	//apply before transform...
			//our matrix lookup
			float rowi = statetci.x + statetci.y * DIM;
			float row = (rowi + .5) / DIMSQ;
			//matrix color
			vec2 weighttc = vec2(row,col);	//its symmetric so col/row swap doesn't matter
			vec4 weightcolor = texture2D(weights, weighttc);
			weightcolor = weightcolor;
			accum += statecolor * weightcolor;
		}
	}
	//euler integration of our derivative
	accum = mix(src, accum, dt);
	//apply after transform?
	accum = signal(accum);
	//remap from [-1,1] to [0,1] for displaying's sake
	gl_FragColor = accum * .5 + .5;
}
]];
}
ShaderProgram.useNone()

--[[
now that we have the hopfield stuff ...
feed in a sample image and watch it settle?
yeah ... so pick a dealio...
and process it through the system
by turning into a vector (unravel)
and then doing a matrix-multiplication on it
...which means using a custom for-loop shader, or a reduction shader
--]]
function update()
	state:swap()
	state:draw{
		viewport={0, 0, sampleDim, sampleDim};
		resetProjection=true;
		texs={weights:cur(), state:prev()};
		shader=convergeShader;
		uniforms={dt=timestep};
	}
end

function drawQuad()
	gl.glBegin(gl.GL_QUADS)
	gl.glTexCoord(0,1) gl.glVertex(0,0)
	gl.glTexCoord(1,1) gl.glVertex(1,0)
	gl.glTexCoord(1,0) gl.glVertex(1,1)
	gl.glTexCoord(0,0) gl.glVertex(0,1)
	gl.glEnd()
end

local mouselastpos = {0,0}
local pan = {0,0}
local scale = -2.30258509299
function glush.display()
	glreport('display begin')
	
	local ar
	do
		local vp = gl.glGet(gl.GL_VIEWPORT)
		ar = vp[3] / vp[4]
	end
	
	gl.glMatrixMode(gl.GL_PROJECTION)
	gl.glLoadIdentity()
	gl.glOrtho( -ar/2, ar/2, -1/2, 1/2, -1,1)
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glLoadIdentity()
	do
		local es = math.exp(scale)
		gl.glScale(es,es,1)
	end
	gl.glTranslate(pan[1], pan[2], 0)
	
	do
		local x,y,b = glush.mouse()
		local dx = x - mouselastpos[1]
		local dy = y - mouselastpos[2]
		if b == 1 then
			-- inverse transform from x,1-y to our screen area
			-- time to use modelview in a shader maybe?
			local mx = x
			local my = y
			-- map to min=[(1-ar)/2,0] max=[(ar+1)/2, 1]
			mx = (mx * ar - ar/2)
			my = my * 1 - 1/2
			local es = math.exp(scale)
			mx = mx / es		my = my / es
			mx = mx - pan[1]	my = my - pan[2]
			mx = mx + 4
			mx = mx / 4			my = my / 4
			if mx >= 0 and mx <= 1 and my >= 0 and my <= 1 then

				-- then state fbo update a dot ...
				state:swap()
				state:draw{
					viewport={0,0,sampleDim,sampleDim};
					resetProjection=true;
					color={1,1,1};
					callback=function()
						gl.glEnable(gl.GL_TEXTURE_2D)
						state:prev():bind()
						drawScreenQuad()
						Tex2D:unbind()
						
						gl.glPointSize(pointSize)
						if not erase then
							gl.glColor(1,1,1)
						else
							gl.glColor(0,0,0)
						end
						gl.glBegin(gl.GL_POINTS)
						gl.glVertex(mx,1-my)
						gl.glEnd()
						gl.glPointSize(1)
					end;
				}
			else
				pan[1] = pan[1] + dx * ar * math.exp(-scale)
				pan[2] = pan[2] + dy * math.exp(-scale)
			end
		end
		if b == 2 then
			scale = scale + dy
		end
		mouselastpos[1] = x
		mouselastpos[2] = y
	end

	if doReset then
		resetState()
		doReset = false
	end
	
	gl.glColor(1,1,1)
	gl.glEnable(gl.GL_TEXTURE_2D)

	-- draw the state
	state:cur():bind()
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glPushMatrix()
	gl.glTranslate(-4,0,0)
	gl.glScale(4,4,1)
	drawQuad()
	Tex2D:unbind()
	
	gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
	drawQuad()
	gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
	
	gl.glPopMatrix()
	
	-- draw the weights 
	weights:cur():bind()
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glPushMatrix()
	gl.glTranslate(-3,-3,0)
	gl.glScale(2,2,1)
	drawQuad()
	gl.glPopMatrix()

	-- draw its inputs
	do
		local w = math.ceil(math.sqrt(#sampleTexs))
		local h = math.ceil(#sampleTexs / w)
		local index = 1
		for i=1,w do
			for j=1,h do
				local v = sampleTexs[index]
				gl.glMatrixMode(gl.GL_MODELVIEW)
				gl.glPushMatrix()

				gl.glTranslate(i*1.1,(h/2-j*1.1),0)
				v.sample:bind()
				drawQuad()
				Tex2D:unbind()
	
				gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
				drawQuad()
				gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
				
--				gl.glTranslate(-1.1,0,0)
--				v.full:bind()
--				drawQuad()
				
				gl.glPopMatrix()
				
				index=index+1
				if index > #sampleTexs then break end
			end
		end
	end

	Tex2D:unbind()
	gl.glDisable(gl.GL_TEXTURE_2D)

	if doUpdate or doSingleUpdate then
		if doSingleUpdate then
			doUpdate = false
			doSingleUpdate = false
		end
		update()
	end
	glreport('display end')
end

if glush.run then glush.run() end

--[[ if not using a display function...
repeat
	update()
	print(l:tostring())
until false
--]]

